(ns agent
  (:require-macros [cljs.core.async.macros :refer [go]])
  (:require [cljs.nodejs :as node]
            [cljs.core.async :refer [<!]]
            [cljs.core.async.interop :refer-macros [<p!]]
            [client :as client]
            [auth :as auth]
            [printer :as printer]
            [spinner :as spinner]
            [clojure.string :as cljstr]))

(def crypto (node/require "crypto"))
(def inquirer (node/require "inquirer"))
(def process (node/require "child_process"))

(defn- valid-event? [connect-time]
  ;; connect-time format: YYYY/MM/DD HH:MM
  ;; we coerce to UTC because the API returns this format in UTC
  (let [d (js/Date. (str connect-time "Z"))
        now (js/Date.)
        ;; add 15 days to connect time
        _ (.setDate d (+ (.getDate d) 15))]
    (> d now)))

(defn- connect-time->relative [connect-time]
  ;; coerce to UTC
  (let [d (.getTime (js/Date. (str connect-time "Z")))
        now (.getTime (js/Date.))
        diff (.abs js/Math (- d now))]
    (cond
      (> diff (* 1000 60 60 24)) ;; days
      (str (js/parseInt (/ diff (* 1000 60 60 24)) 10) "d")
      (> diff (* 1000 60 60)) ;; hours
      (str (js/parseInt (/ diff (* 1000 60 60)) 10) "h")
      (> diff (* 1000 60)) ;; minutes
      (str (js/parseInt (/ diff (* 1000 60)) 10) "m")
      :else
      (str (js/parseInt (/ diff 1000) 10) "s"))))

(defn- parse-response [{:keys [fn-callback] :as result}]
  (go (when-let [response (<! (:response-chan result))]
        (some-> response
                spinner/stop
                client/extract-body
                clj->js
                fn-callback))))

;; agent create token

(declare create-token
         ask-to-continue
         update-org)

(defn create-agent-token [command]
  (if (.-force command)
    (create-token)
    (ask-to-continue)))

(defn- create-token []
  (some-> {}
          auth/key-auth
          update-org
          parse-response))

(defn- ask-to-continue []
  (go
    (let [message "If you have agents running in production, this operation will break then. Do you wish to continue?\n(1) Yes\n(2) No\n->"
          prompt (-> inquirer
                     (.prompt (clj->js [{:type "input" :name "create-token" :message message}])))
          response (get (js->clj (<p! prompt)) "create-token")]
      (if (= "1" response)
        (create-token)
        (js/console.log "aborting ...")))))


(defn- update-org [task]
  (let [token (-> crypto
                  (.randomBytes 24)
                  (.toString "hex"))
        _ (spinner/start nil)
        token-sha256 (-> crypto
                         (.createHash "sha256")
                         (.update token)
                         (.digest "hex"))
        response-chan (client/update-agent-token (assoc task :agent-token token-sha256))]
    (assoc task :response-chan response-chan
           :agent-token token
           :fn-callback (fn [_] (js/console.log token)))))

;; list agents

(declare list-agents-api)

(defn list-agents [_]
  (some-> {}
          auth/key-auth
          list-agents-api
          parse-response))

(defn- parse-outcome [task]
  (if (empty? task)
    task
    (let [events (js->clj task)
          events-up (filter #(= (get % "status") "UP") events)
          events-down (filter #(and (valid-event? (get % "connect_time"))
                                    (= (get % "status") "DOWN")) events)
          events (concat events-up events-down)]
      (some-> (map #(vector (get % "tags")
                            (get % "app_version")
                            (get % "hostname")
                            (connect-time->relative
                             (get % "connect_time"))
                            (get % "status"))
                   events)
              (#(conj {} {:content %
                          :outcome "Agents Connected"
                          :print-type :list
                          :table-head ["Tag" "Version" "Hostname" "Connected-At" "Status"]
                          :columns-width [25, 10, 40, 15, 10]}))
              printer/print-data))))

(defn- list-agents-api [task]
  (let [_ (spinner/start nil)
        response-chan (client/list-agents task)]
    (assoc task :response-chan response-chan
           :fn-callback parse-outcome)))

;; docker-run

(defn- shell [{:keys [command args stdout-cb stderr-cb dry-run]}]
  (if dry-run
    (js/console.log command (cljstr/join " " (clj->js args)))
    (let [_ (js/console.log command (cljstr/join " " (clj->js args)))
          p (-> process
                (.spawn command (clj->js args)))]
      (-> p (.-stdout)
          (.on "data" stdout-cb))
      (-> p (.-stderr)
          (.on "data" stderr-cb))
      nil)))

(defn docker-run [command]
  (let [agent-token (.-token command)
        tags (.-tags command)
        ;; docker env list
        env-list (concat (.-env command)
                         (when-not (cljstr/blank? agent-token)
                           [(str "TOKEN=" agent-token)])
                         (when-not (cljstr/blank? tags)
                           [(str "TAGS=" tags)]))
        ;; runops env-var
        env-var-name (.-envvar command)
        env-var-list (.-ev command)
        env-var-map (into {} (map #(clojure.string/split % "=") env-var-list))
        env-var-json (.stringify js/JSON (clj->js env-var-map))

        rm? (.-rm command)
        container-name (.-name command)
        _ (shell {:dry-run (.-dryrun command)
                  :command "docker"
                  :args (concat ["run"]
                                (when rm? ["--rm"])
                                (when-not (cljstr/blank? container-name)
                                  [(str "--name=" container-name)])
                                (when-not (empty? env-var-list)
                                  [(str "-e=" env-var-name "=" env-var-json)])
                                (when-not (empty? env-list)
                                  (map #(str "-e=" %) env-list))
                                ["runops/agent:latest"])
                  :stdout-cb (fn [data] (-> js/process (.-stdout) (.write (.toString data))))
                  :stderr-cb (fn [data] (-> js/process (.-stdout) (.write (.toString data))))})]))

(comment
  (docker-run (clj->js {:token "foo" :dryrun true :tags "local" :name "agent" :rm  true
                        :envvar "ENV_CONF"
                        :ev ["PG_HOST=127.0.0.1" "PG_DB=brutz" "PG_USER=root" "PG_PASS=passme"]
                        :env ["AWS_SECRET_MANAGER_DISABLE=true" "JWK_URL=https://api.runops.io/jwk-url"]})))