(ns client
  (:require [cljs-http.client :as http]
            [cljs.reader :refer [read-string]]
            [goog.crypt.base64 :as base64]
            [clojure.string :as str]))

(def api-url "https://api.runops.io")

(defn extract-body
  "Returns the `:body` key of the `response`.
  It returns the `:body` key of the map for successful HTTP status codes.
  Error HTTP status codes displays the `:message` key to the user,
  and return `nil`."
  [response]
  (case (:status response)
    200 (:body response)
    201 (:body response)
    0 (js/console.log "Failed to connect to the server. Please check your network connection and try again.")
    (js/console.log (:message (:body response)))))

(defn authorization-value [key]
  (if (str/starts-with? key "Bearer")
    key
    (str "Basic " (base64/encodeString (str key ":")))))

(defn base-headers
  ([key]
   {:headers {"Authorization" (authorization-value key)
              "Accept" "application/edn"
              "User-Agent" "runops/cli"}
    :with-credentials? false})
  ([key ua]
   {:headers {"Authorization" (authorization-value key)
              "Accept" "application/edn"
              "User-Agent" (if (= (read-string ua) nil) "runops/cli" ua)}
    :with-credentials? false}))

;;;; Tasks

(defn get-task [{:keys [id key]}]
  (http/get (str api-url "/v1/tasks/" id)
            (base-headers key)))

(defn run-task [{:keys [id key]}]
  (http/post (str api-url "/v1/tasks/" id "/runs")
             (base-headers key)))

(defn kill-task [{:keys [id key]}]
  (http/post (str api-url "/v1/tasks/" id "/kill")
             (base-headers key)))

(defn get-task-logs [{:keys [id key]}]
  (http/get (str api-url "/v1/tasks/" id "/logs")
            (base-headers key)))

(defn list-tasks [{:keys [key]}]
  (http/get (str api-url "/v1/tasks")
            (base-headers key)))

(defn create-task [{:keys [key user-agent message type target script]}]
  (http/post (str api-url "/v1/tasks")
             (conj (base-headers key user-agent)
                   {:json-params {:message message
                                  :type type
                                  :target target
                                  :script script}})))

;;;; Targets 

(defn get-target [{:keys [key name]}]
  (http/get (str api-url "/v1/targets/" name)
            (base-headers key)))

(defn list-targets [{:keys [key]}]
  (http/get (str api-url "/v1/targets")
            (base-headers key)))

(defn create-target [target]
  (http/post (str api-url "/v1/targets")
             (conj (base-headers (:key target))
                   {:json-params (dissoc target :key)})))

(defn update-target [target]
  (http/put (str api-url "/v1/targets")
            (conj (base-headers (:key target))
                  {:json-params (dissoc target :key)})))

;;;; Templates

(defn list-templates [{:keys [key]}]
  (http/get (str api-url "/v1/templates")
            (base-headers key)))

(defn run-template [template]
  (http/post (str api-url "/v1/templates/runs")
             (conj (base-headers (:key template))
                   {:json-params (dissoc template :key)})))

(defn get-template [{:keys [name key]}]
  (http/get (str api-url "/v1/templates/" name)
            (base-headers key)))

;;;; Users

(defn list-users [{:keys [key]}]
  (http/get (str api-url "/v1/users")
            (base-headers key)))

(defn create-user [user]
  (http/post (str api-url "/v1/users")
             (conj (base-headers (:key user))
                   {:json-params (dissoc user :key)})))

(defn update-user [user]
  (http/put (str api-url "/v1/users/" (:id user))
            (conj (base-headers (:key user))
                  {:json-params (dissoc user :id :key)})))

;;;; Org

(defn update-agent-token [{:keys [key agent-token]}]
  (http/put (str api-url "/v1/orgs")
            (conj (base-headers key)
                  {:json-params {:sha_256_agents_token agent-token}})))

(defn list-agents [{:keys [key]}]
  (http/get (str api-url "/v1/agents")
            (base-headers key)))

;;;; Version

(defn get-npm-version []
  (http/get "https://registry.npmjs.org/runops/latest"
            {:timeout 2000}))
