(ns targets
  (:require [cljs.core.async :refer [<! go]]
            [client :as client]
            [traces :as traces]
            [auth :as auth]
            [spinner :as spinner]
            [printer :as printer]
            [clojure.string :as str]))

(defn parse-groups [target]
  (if-let [groups (:groups target)]
    (let [groups (str/split groups #",")]
      (assoc target :groups groups))
    target))

(defmulti targets-values (fn [data] (map? data)))

(defmethod targets-values false [data]
  (->> data
       (map #(vals (select-keys % [:name :type :review_type :reviewers :tags])))))

(defmethod targets-values true [data]
  data)

(defn parse-response [result]
  (go (when-let [response (<! (:response-chan result))]
        (some-> response
                spinner/stop
                client/extract-body
                targets-values
                (#(conj {} {:content %
                            :outcome (:outcome result)
                            :print-type (:print-type result)
                            :table-head (:table-head result)
                            :columns-width (:columns-width result)}))
                printer/print-data
                traces/close))))

(defn key-value-str->map [data]
  (let [index (str/index-of data "=" 0)]
    {:name (subs data 0 index)
     :value (subs data (+ index 1))}))

;; create target
(defn create-target-api [target]
  (let [_ (spinner/start nil)
        response-chan (client/create-target target)]
    (assoc nil :response-chan response-chan)))

(defn create-outcome [data]
  (assoc data :print-type :get
         :outcome "Target created successfully !!"))

(defn create [command]
  (.withTrace traces/event (clj->js {:command "create-target"})
              #(some-> (conj {:name (.-name command)
                              :type (.-type command)}
                             (when-let [tags (.-tags command)]
                               {:tags tags})
                             (when-let [message (.-message command)]
                               {:message message})
                             (when-let [groups (.-groups command)]
                               {:groups groups})
                             (when-let [review_type (.-review command)]
                               {:review_type review_type})
                             (when-let [reviewers (.-reviewers command)]
                               {:reviewers reviewers})
                             (when-let [channel_name (.-channel command)]
                               {:channel_name channel_name})
                             (when-let [secret_path (.-secret_path command)]
                               {:secret_path secret_path})
                             (when-let [secret_provider (.-secret_provider command)]
                               {:secret_provider secret_provider})
                             (when-let [secret_mapping (.-secret_mapping command)]
                               {:secret_mapping secret_mapping})
                             (when-let [config (.-config command)]
                               {:config config})
                             (when-let [redact (.-redact command)]
                               {:redact redact}))
                       parse-groups
                       auth/key-auth
                       create-target-api
                       create-outcome
                       parse-response)))

;; update target
(defn update-target-api [target]
  (let [_ (spinner/start nil)
        response-chan (client/update-target target)]
    (assoc nil :response-chan response-chan)))

(defn update-outcome [data]
  (assoc data :print-type :get
         :outcome "Target updated successfully !!"))

(defn update-cmd [command]
  (.withTrace traces/event (clj->js {:command "update-target"})
              #(some-> (conj {:name (.-name command)}
                             (when-let [type (.-type command)]
                               {:type type})
                             (when-let [message (.-message command)]
                               {:message message})
                             (when-let [groups (.-groups command)]
                               {:groups groups})
                             (when-let [review (.-review command)]
                               {:review_type review})
                             (when-let [reviewers (.-reviewers command)]
                               {:reviewers reviewers})
                             (when-let [channel (.-channel command)]
                               {:channel_name channel})
                             (when-let [secret_provider (.-secret_provider command)]
                               {:secret_provider secret_provider})
                             (when-let [secret_path (.-secret_path command)]
                               {:secret_path secret_path})
                             (when-let [secret_mapping (.-secret_mapping command)]
                               {:secret_mapping secret_mapping})
                             (when-let [config (.-config command)]
                               {:config config})
                             (when-let [tags (.-tags command)]
                               {:tags tags})
                             (when-let [redact (.-redact command)]
                               {:redact redact})
                             (when-let [status (.-status command)]
                               {:status status}))
                       parse-groups
                       auth/key-auth
                       update-target-api
                       update-outcome
                       parse-response)))

;; list target
(defn list-targets-api [target]
  (let [_ (spinner/start nil)
        response-chan (client/list-targets target)]
    (assoc target :response-chan response-chan)))

(defn list-outcome [data]
  (assoc data :print-type :list
         :outcome "List targets"
         :table-head ["Name" "Type" "Review" "Reviewers" "Tags"]
         :columns-width [25, 15, 10, 20, 10]))

(defn list-targets []
  (.withTrace traces/event
              (clj->js {:command "list-targets"})
              #(some-> (auth/key-auth {})
                       list-targets-api
                       list-outcome
                       parse-response)))

;; get target
(defn get-target-api [target]
  (let [_ (spinner/start nil)
        response-chan (client/get-target target)]
    (assoc target :response-chan response-chan)))

(defn get-outcome [data]
  (assoc data :print-type :get
         :outcome "Get target"))

(defn get-target [command]
  (.withTrace traces/event
              (clj->js {:command "get-target"})
              #(some-> (auth/key-auth {:name (.-name command)})
                       get-target-api
                       get-outcome
                       parse-response)))
