(ns tasksrepl
  (:require [cljs.nodejs :as node]
            [cljs.core.async :refer [<! go]]
            [client :as client]
            [traces :as traces]
            [auth :as auth]
            [spinner :as spinner]
            [clojure.string :as str]
            [tasks :as tasks]))

(def readline (node/require "readline"))

(def help-message "
  COMMANDS
     
     :help issue this help
     :target <name> set a target to run commands
  
  HOTKEYS

     ctrl+e toogle to multiline prompt, on second toogle evaluates the command
     ctrl+c cancel the multiline evaluation or exits REPL
  
  USAGE

     Each command executed is processed straight from REPL and processed remotely.
     This is the default mode, pressing ctrl+e will toogle to the multiline which
     could be evaluated pressing ctrl+e again.
     The :target command will allow selecting a specific target to run a command.
     
     A complete guide is available at https://runops.io/docs/user-guides/REPL")

(def welcome-message
  "REPL started, each command will be executed remotely
and displayed in this session. https://runops.io/docs/user-guides/REPL
Type :help to list available commands
")

(defn- set-repl-state [state key val]
  (swap! state assoc (keyword key) val))

(defn- set-repl-default-state
  "set repl to default action and clear the repl input"
  [state]
  (swap! state assoc :action :toogle-single-line-prompt
         :repl-input nil))

(defn- set-repl-prompt
  "sets the prompt of repl to multiline
  when repl arg is passed. set to single line
  when repl and state is passed."
  ([repl]
   (.setPrompt repl "#_=> ")
   (.prompt repl))
  ([repl state]
   ;; in this mode, always clear the input and set to default action
   (set-repl-default-state state)
   (.setPrompt repl (str (@state :target) "=> "))
   (.prompt repl)))

(defn- process-api-response [response]
  (case (response :status)
    200 (do (spinner/stop-repl nil)
            (:body response))
    201 (do (spinner/stop-repl nil)
            (:body response))
    0 (do (spinner/stop-repl {:spinner-msg "failed to connect to the server!"
                              :spinner-action :fail})
          {:err true
           :message (str "Failed to connect to the server,\n "
                         "please check your network connection and try again")})
    (do
      (spinner/stop-repl {:spinner-msg "failed executing task!"
                          :spinner-action :fail})
      {:err true
       :message (:message (:body response))})))

(defn- process-api-error [response]
  (if (true? (response :err))
    (do (js/console.log "error:" (response :message))
        (set-repl-prompt (response :repl) (response :repl-state))
        nil)
    response))

(defn- process-task-response [result]
  (go (when-let [response (<! (:response-chan result))]
        (some-> response
                traces/flush-event
                process-api-response
                (assoc :repl-state (result :repl-state)
                       :repl (result :repl))
                process-api-error
                (#(if (or (:task_logs %) (:logs_url %))
                    (do (tasks/parse-logs-response %) %)
                    (do (tasks/parse-task-no-logs (dissoc % :repl
                                                          :target
                                                          :repl-state) result) %)))
                (#(set-repl-prompt (:repl %) (:repl-state %)))))))

(defn- process-task-request [task]
  (spinner/start-repl {:spinner-msg "executing task remotely ..."})
  (assoc task
         :response-chan (client/create-task (assoc task
                                                   :user-agent "runops/cli+repl"))
         :print-type :get))

(defn- create-task-repl [repl state name data]
  (.withTrace traces/event (clj->js {:command "create-task"})
              #(some-> (conj {:target name
                              :script data
                              :repl repl
                              :repl-state state})
                       auth/key-auth
                       process-task-request
                       process-task-response)))

(defn- process-repl-commands [data]
  ;; don't process commands if it's not in this state
  (let [line (data :line)
        repl (data :repl)
        state (data :state)]
    (when (not (= (@state :action) :toogle-single-line-prompt))
      data)
    (cond
      (re-matches #"^:help$" line)
      (do
        (js/console.log help-message)
        (set-repl-prompt repl state)
        nil)
      (re-matches #":target\s[\w+-\.:]+$" line)
      (do
        (let [target-name (str/replace line ":target " "")]
          (set-repl-state (data :state) :target target-name)
          (set-repl-prompt repl state))
        nil)
      :else data)))

(defn- process-repl-actions [data]
  (let [line (data :line)
        repl (data :repl)
        state (data :state)]
    (case (@state :action)
      :toogle-multi-line-prompt
      (do
        (set-repl-prompt repl)
        ;; it will preserve empty lines
        (set-repl-state state :repl-input (str
                                           (@state :repl-input) "\n"
                                           line))
        data)
      :run-multi-line-prompt
      (do
        (if (empty? (get @state :target))
          (js/console.log "no target set, run :target <name>")
          (if (= (str/trim-newline (@state :repl-input)) "")
            ()
            (create-task-repl
             repl state (@state :target) (@state :repl-input))))
        (set-repl-default-state state)
        (set-repl-prompt repl state)
        data)
      :toogle-single-line-prompt
      (do
        (cond
          (= line "") (set-repl-prompt repl state)
          (empty? (get @state :target)) (js/console.log "no target set, run :target <name>")
          :else
          (do
            (create-task-repl repl state (@state :target) line)
            (set-repl-state state :action :task-running)))
        data)
      nil)))

(defn- toogle-multi-line-prompt [state repl]
  (if (= (@state :action) :toogle-multi-line-prompt)
    (set-repl-state state :action :run-multi-line-prompt)
    (do
      (set-repl-state state :action :toogle-multi-line-prompt)
      (js/console.log "")
      (set-repl-prompt repl))))

(defn start-repl [welcome-message]
  (.setRawMode node/process.stdin true)
  (.emitKeypressEvents readline node/process.stdin)
  (let [repl (.createInterface readline (clj->js {:input node/process.stdin
                                                  :output node/process.stdout}))
        state (atom {:action :toogle-single-line-prompt})]
    (when welcome-message
      (js/console.log welcome-message))
    (set-repl-prompt repl state)
    (.on node/process.stdin "keypress"
         (fn [_ key]
           (let [name (get (js->clj key) "name")
                 ctrl (get (js->clj key) "ctrl")]
             (when (and (true? ctrl) (= name "e"))
               (toogle-multi-line-prompt state repl)))))
    (.on repl "line"
         (fn [line]
           (some-> (conj {:line line
                          :state state
                          :repl repl})
                   process-repl-commands
                   process-repl-actions)))
    (.on repl "close"
         (fn []
           (case (@state :action)
             :toogle-multi-line-prompt (start-repl nil)
             ;; stop the spinner and leave the task in the "background"
             :task-running (do (spinner/stop-repl {:spinner-action :info
                                                   :spinner-msg
                                                   "the task may still be running remotely"})
                               (js/console.log "")
                               (.exit node/process 0))
             (do
               (js/console.log "")
               (.exit node/process 0)))))))
