(ns templates
  (:require [cljs.core.async :refer [<! go]]
            [client :as client]
            [traces :as traces]
            [auth :as auth]
            [spinner :as spinner]
            [clojure.string :as str]
            [printer :as printer]))

(defmulti templates-values (fn [data] (map? data)))

(defmethod templates-values false [data]
  (->> data
       (map #(assoc {} :name %))
       (map #(vals (select-keys % [:name])))))

(defmethod templates-values true [template]
  (if (:params template)
    (assoc template :params (str/join "\n" (:params template)))
    template))

(defn parse-task-no-logs [task result]
  (some-> task
          templates-values
          (#(conj {} {:content %
                      :outcome (:outcome result)
                      :print-type (:print-type result)
                      :table-head (:table-head result)
                      :columns-width (:columns-width result)}))
          printer/print-data))

(defn extract-logs [task]
  (if (:task_logs task) (:task_logs task) task))

(defn parse-logs-response [task]
  (some-> task
          extract-logs
          clj->js
          js/console.log))

(defn parse-response [result]
  (go (when-let [response (<! (:response-chan result))]
        (some-> response
                spinner/stop
                client/extract-body
                (#(if (or (:task_logs %) (:logs_url %))
                    (parse-logs-response %)
                    (parse-task-no-logs % result)))
                traces/close))))

;; list templates
(defn list-templates-api [user]
  (let [_ (spinner/start nil)
        response-chan (client/list-templates user)]
    (assoc user :response-chan response-chan)))

(defn list-outcome [data]
  (assoc data :print-type :list
         :outcome "List templates"
         :table-head ["Name"]
         :columns-width [40]))

(defn list-templates []
  (.withTrace traces/event
              (clj->js {:command "list-templates"})
              #(some-> (auth/key-auth {})
                       list-templates-api
                       list-outcome
                       parse-response)))

;; get template
(defn get-template-api [template]
  (let [_ (spinner/start nil)
        response-chan (client/get-template template)]
    (assoc template :response-chan response-chan)))

(defn get-outcome [data]
  (assoc data :print-type :get
         :outcome "Get template"))

(defn get-template [command]
  (.withTrace traces/event
              (clj->js {:command "get-template"})
              #(some-> (auth/key-auth {:name (.-name command)})
                       get-template-api
                       get-outcome
                       parse-response)))

;; run template
(defn run-template-api [template]
  (let [_ (spinner/start nil)
        response-chan (client/run-template template)]
    (assoc template :response-chan response-chan)))

(defn run-outcome [data]
  (assoc data :print-type :get
         :outcome "Run template !!"))

(defn params-str->map [params param]
  (let [index (str/index-of param "=" 0)]
    (conj params
          {(keyword (subs param 0 index)) (subs param (+ index 1))})))

(defn parse-params [template]
  (if-let [params (:params template)]
    (let [params (str/split params #",")
          params (reduce params-str->map {} params)]
      (assoc template :params params))
    template))

(defn run-template [run]
  (.withTrace traces/event
              (clj->js {:command "run-template"})
              #(some-> (conj {:name (.-name run)
                              :target (.-target run)}
                             (when-let [message (.-message run)]
                               {:message message})
                             (when-let [task-type (.-type run)]
                               {:type task-type})
                             (when-let [params (.-params run)]
                               {:params params}))
                       parse-params
                       auth/key-auth
                       run-template-api
                       run-outcome
                       parse-response)))
