(ns users
  (:require [cljs.core.async :refer [<! go]]
            [client :as client]
            [traces :as traces]
            [auth :as auth]
            [spinner :as spinner]
            [printer :as printer]
            [clojure.string :as str]))

(defn parse-groups [user]
  (if-let [groups (:groups user)]
    (let [groups (str/split groups #",")]
      (assoc user :groups groups))
    user))

(defmulti users-values (fn [data] (map? data)))

(defmethod users-values false [data]
  (->> data
       (map #(vals (select-keys % [:id :name :email :status :team :groups])) data)))

(defmethod users-values true [data]
  data)

(defn parse-response [result]
  (go (when-let [response (<! (:response-chan result))]
        (some-> response
                spinner/stop
                client/extract-body
                users-values
                (#(conj {} {:content %
                            :outcome (:outcome result)
                            :print-type (:print-type result)
                            :table-head (:table-head result)
                            :columns-width (:columns-width result)}))
                printer/print-data
                traces/close))))

;; create user
(defn create-user-api [user]
  (let [_ (spinner/start nil)
        response-chan (client/create-user user)]
    (assoc user :response-chan response-chan)))

(defn create-outcome [data]
  (assoc data :print-type :get
         :outcome "User created successfully !!"))

(defn create [command]
  (.withTrace traces/event (clj->js {:command "create-user"})
              #(some-> (conj {:name (.-name command)
                              :email (.-email command)}
                             (when-let [status (.-status command)]
                               {:status status})
                             (when-let [team (.-team command)]
                               {:team team})
                             (when-let [groups (.-groups command)]
                               {:groups groups}))
                       parse-groups
                       auth/key-auth
                       create-user-api
                       create-outcome
                       parse-response)))

;; update user
(defn update-user-api [user]
  (let [_ (spinner/start nil)
        response-chan (client/update-user user)]
    (assoc user :response-chan response-chan)))

(defn update-outcome [data]
  (assoc data :print-type :get
         :outcome "User updated successfully !!"))

(defn update-cmd [command]
  (.withTrace traces/event (clj->js {:command "update-user"})
              #(some-> (conj {:id (.-id command)}
                             (when-let [status (.-status command)]
                               {:status status})
                             (when-let [team (.-team command)]
                               {:team team})
                             (when-let [groups (.-groups command)]
                               {:groups groups}))
                       parse-groups
                       auth/key-auth
                       update-user-api
                       update-outcome
                       parse-response)))

;; list User
(defn list-users-api [users]
  (let [_ (spinner/start nil)
        response-chan (client/list-users users)]
    (assoc users :response-chan response-chan)))

(defn list-outcome [data]
  (assoc data :print-type :list
         :outcome "List users"
         :table-head ["Id" "Name" "Email" "Status" "Team" "Groups"]
         :columns-width [7, 10, 23, 10, 10, 20]))

(defn list-users []
  (.withTrace traces/event
              (clj->js {:command "list-users"})
              #(some-> (auth/key-auth {})
                       list-users-api
                       list-outcome
                       parse-response)))
